/*
 * Copyright (C) 2004 Andres Sajo
 *
 * This file is part of the nisperod package.
 *
 * nisperod is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
 *
 *
 * 1wire library by Soros Jozsef with code portions by Justin Huff.
 * 
 *
 *
 * $Id: main.c,v 1.17 2004/10/08 14:31:24 andres Exp andres $ 
 */

/*
 * TODO:
 *
 * Urgency: normal
 * 
 * Second client never get to pass the Bus.Reset() call.
 *
 *
 * Urgency: Normal
 * 
 * Write to --pid-file file the pid of the detached process, so start-stop-daemon 
 * can gracefully stop this daemon, and not statically from PIDFILE.
 *
 */



/* Code portions are from:
 * 
 * uszerver.c - Egyszeru pelda szerver a Unix domain socket 
 * hasznalatara. 
 *
 * REF[1]: modified code from: LINUX Programaza's, Ba'nya'sz Ga'bor - Levendovszky Tihame'r. 2003 SZAK kiado
 * ISBN 963 9131 57 1
 *
 * REF[2]: this code comes from : Advanced programming in the UNIX(R) Environment
 * W. Richard Stevens, Addisson-Wesley Longman Inc.
 * ISBN 0201563177, Chapter 13, Daemon Processes
 *
 *
 *
 *  ISSUES:
 *    1   in unixds.c: 
 *        nis_conn = accept(nis_sock, (struct sockaddr *) &nis_address, &nis_addrLength);
 *        call blocks the progress, man 2 accepts says:
 *
 *    If no pending connections are present on the queue, and the socket is not marked as
 *    non-blocking, accept blocks the caller until  a  connection  is  present.   If  the
 *    socket  is marked non-blocking and no pending connections are present on the queue,
 *    accept returns EAGAIN.
 *
 *        it just dosen't say where to set the non-blocking mark.
 *
 *    2   in wrappers.cpp:
 *        bus.Reset() on second client gets a FAILED return.
 *
 * SOLUTIONS:
 *    1   is resolved by adding to main.c line 
 *        if ( fcntl(nis_sock, F_SETFL, O_NONBLOCK) < 0 ){
 *            fprintf(stderr, "%s: %s\n", name, sys_errlist[errno]);
 *           term = TRUE;
 *           return EXIT_FAILURE;
 *        }
 */

#include <fcntl.h>
#include <signal.h>
#include <sys/poll.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <sys/ipc.h>
#include <sys/sem.h>

#include "nisperod.h"
int parse_argv(int argc, char *argv[], const char *name);

struct child_pid_t * chld_st = NULL;
/* volatile int chldcount = 0; */


int child_prn_list(void);
int semid = 0;

int main(int argc, char *argv[]) {
	const char version[] = {"$Id: main.c,v 1.17 2004/10/08 14:31:24 andres Exp andres $"};
	const char revision[] = {"$Revision: 1.17 $"};
	const char name[] = {"nisperod"};
	char table_path[] = {TABLE};
	char config_path[] = {CONF};
#if 0
	struct def_tab_t * def = NULL;
	struct nispero_t * nis = NULL;
#endif
	int i = 0, ret = 0;

	pid_t pid;

	union semun {
		int val;
		struct semid_ds *buf;
		ushort * array;
	} semarg;


	struct pollfd psock; /* for the poll function */
	FILE * fd = NULL;

	pid = 0; /* getpid(); */

	if (sizeof(unsigned long long) != 8) {
		fprintf(stderr, "%s: type \'unsigned long long\' is not 64 bit long.\n"
				"(Note that long long is not an ANSI C type. Any\n"
				"program using this will not be portable to all architectures).\n",
				name);
		return EXIT_FAILURE;
	}

	/* put some hardcoded defaults: */
	usercount = 0;
	strncpy(info.port, DEFAULT_PORT, strlen(DEFAULT_PORT));
	strncpy(info.socket, DEFAULT_SOCKET, strlen(DEFAULT_SOCKET));
	/* info.socketcount = DEFAULT_SOCK_CNT; */
	info.verbose = 0; /* no verbose output, can be changed by -V --verbose */
	/* chldcount = 0; */
	
#ifndef TEST /* don't write pid file, because we are usualy not root */
	fd = fopen(PIDFILE, "w");
	if (fd == NULL){
		fprintf(stderr, "%s: error creating pidfile: %s\n",
							 name, sys_errlist[errno]);
		return EXIT_FAILURE;
	}
#endif
		
	if (parse_argv(argc, argv, name) != 0){
		return EXIT_FAILURE;
	}
	
	if (info.verbose > 0) fprintf(stderr, "%s: %s\n", name, version);
	
	
	if (parse_config(config_path) != 0){
		fprintf(stderr, "%s: error while parsing config file \'%s\'.\n",
							 name, config_path);
		return EXIT_FAILURE;
	}
	if (parse_table(table_path) != 0){
		fprintf(stderr, "%s: error while parsing mapping table file \'%s\'.\n",
							 name, table_path);
		return EXIT_FAILURE;
	}
#if 0
	/* this goes to extended protocol 'get tables' or something similar */
	printf("auth table with %i entries:\n", usercount+1);
	while (i <= usercount){
		printf("%i: login = \'%s\', password = \'%s\', protocol = %i\n", i, 
		info.users[i].login, info.users[i].password, info.users[i].protocol);
		i++;
	}
	printf("define table:\n");
	printf("%24s      %22s\n", "define name", "value");
	def = defl_st;
	while (def != NULL){
		printf("\'%32s\' -> \'%18s\'\n", def->name, def->cid);
		def = def->next;
	}

	printf("mapping table:\n");
	printf("%28s       %20s       pins\n", "device attached to", "chip address");
	nis = nisl_st;
	while (nis != NULL){
		printf("\'%32s\' ->  0x%016Lx  :  0x%02x\n",
		nis->name, nis->c_id, nis->outlets);
		nis = nis->next;
	}
#endif

	semid = semget(SEMKEY,1,0666 | IPC_CREAT);
	if(semid < 0) {
		syslog(LOG_ERR, "Unable to obtain semaphore.");
		exit(1);
	}

	semarg.val = 1;
	if(semctl(semid,0,SETVAL,semarg) < 0) {
		syslog(LOG_ERR, "Unable initialize semaphore.");
		exit(1);
	}
	
	if (wrapper_open_dev(info.port) != 0){
		fprintf(stderr, "%s: unable to find 1wire chip(s) on device %s.\n", name, info.port);
		return EXIT_FAILURE;
	}
#if 1
	/* here we cmpare the define table with the c_id found on the 1wire bus: */
	if (wrapper_compare_tables() != 0){
		fprintf(stderr, "%s: define table (%s) does not match \n"
		"devices found on %s.\n", name, TABLE, info.port);
		return EXIT_FAILURE;
	}
#endif
	wrapper_close_dev();

	/* signal handlers */	
	if (signal(SIGTERM, &nispero_term) == SIG_ERR){
		fprintf(stderr, "%s: unable to hook signal handler &nispero_term().\n",
							 name);
		return EXIT_FAILURE;
	}
	if (signal(SIGCHLD, &nispero_wait) == SIG_ERR){
		fprintf(stderr, "%s: unable to hook signal handler &nispero_wait().\n",
							 name);
		return EXIT_FAILURE;
	}
	
	openlog(name, LOG_PID | LOG_CONS, LOG_DAEMON);
	syslog(LOG_INFO, "starting up %s, %s", name, revision);
	syslog(LOG_INFO, "stonith device on \'%s\'", info.port);

	if (open_socket() != 0){
		fprintf(stderr, "%s: error while trying to create the socket.\n", name);
		return EXIT_FAILURE;
	}
	
	syslog(LOG_INFO, "socket: \'%s\' created", info.socket);

	/* from here on, messages go to syslog */
	
	/* NOTE: this might break owfully the code: */
	if ( fcntl(nis_sock, F_SETFL, O_NONBLOCK) < 0 ){
		fprintf(stderr, "%s: %s\n", name, sys_errlist[errno]);
		term = TRUE;
		return EXIT_FAILURE;
	}

	psock.fd = nis_sock;
	psock.events = POLLIN | POLLPRI;
	/* printf("nis_sock descriptor = %i\n", psock.fd); */
	i=0;

		
if (info.verbose == 0){
	/* going to background */
	pid = fork();
	if (pid < 0){
		syslog(LOG_ERR, "error from fork();");
		term = TRUE;
		return EXIT_FAILURE;
	}
	if (pid != 0){
#ifndef TEST
		fprintf(fd, "%i\n", pid);
		fclose(fd);
#endif
		return EXIT_SUCCESS;
	}

	setsid();
	chdir("/");
	umask(0);
	close(0);
	close(1);
	close(2);
}
	syslog(LOG_INFO, "pidfile '%s': %i", PIDFILE, getpid());
	syslog(LOG_INFO, "startup completed.");

cont_polling:
	while ((term == FALSE) && (ret = poll(&psock, (unsigned int)1, (int)-1))){
		
		if (term == TRUE) {
#ifdef DEBUG
			printf("while(poll)->(term==TRUE) breaking\n");
#endif
			break;
		}
		if (ret == -1){
			/* SIGCHLD interrupted it, we can go back to poll. */
			if (errno == SIGILL) goto cont_polling;
			/* other error occured */
			syslog(LOG_ERR, "error from poll()=%i: %s", errno, sys_errlist[errno]);
			term = TRUE;
			continue;
		}
#if 0
		if (chldcount > info.socketcount_deprecated){
			syslog(LOG_WARNING, "max number of connections to server reached, connection refused");
			/* refuse connection */
			/* shutdown((int)info.socket, SHUT_RD); dosen't work */
			
			continue;
		}
#endif
		pid = fork();
		if (pid < 0){
			syslog(LOG_ERR, "error from fork(): %s", sys_errlist[errno]);
			term = TRUE;
			break;
		}
		if (pid == 0){
			syslog(LOG_INFO, "child(%i): handling next connection", getpid());
			if (accept_connect() != 0) {
				syslog(LOG_ERR, "child() exited with error status");
				sleep(3);
				_exit(1);
			}
			/* child has to exit */
			syslog(LOG_INFO, "child(%i): exiting.", getpid());
			_exit(0);
		} else {
			/* increment & add to list of pids */
			/* chldcount++; */
			if (add_pool(pid) != 0){
				abort();
			}
#ifdef DEBUG
			child_prn_list();
#endif
			syslog(LOG_INFO, "new child %i", pid);
		}
		if (term == TRUE) break;
	}
	
	
	
	/* wait for children:
	 * The handler of SIGTERM takes care of the children. 
	 */
	
	syslog(LOG_INFO, "closing and removing socket.");
	close(nis_sock);
	unlink(info.socket);
	/* TODO: static pidfile -> dinamic */
	unlink(PIDFILE);
	syslog(LOG_INFO, "unload complete, good bye.");
	return EXIT_SUCCESS;
}




int parse_argv(int argc, char *argv[], const char *name){
	
	/* We are to use get_opts_long in 
	 * the future but for now there's 
	 * no reason
	 */
	if (argc == 1) return 0;
	if ((argc == 2) && (strcmp(argv[1], "-d") == 0)){
		info.verbose = 1;
		return 0;
	}
	/* print_usage */
	fprintf(stderr, "usage: %s [-d]\n  -d : dosen't detach processes from terminal.\n", name);

	return 1;
	
}



/* some pid handling routines: */

int child_prn_list(void){
	struct child_pid_t * tmp;
	tmp = chld_st;
	printf("pid pool:\n");
	while (tmp != NULL){
		printf("%8p\t->\t%05i\t->\t%8p\n", tmp->prev, tmp->pid, tmp->next);
		tmp = tmp->next;
	}
	return 0;
}

int add_pool(pid_t pid){
	struct child_pid_t * tmp;
	struct child_pid_t * s;
	syslog(LOG_INFO, "adding %i to pool.", pid);
	tmp = (struct child_pid_t *)malloc(sizeof(struct child_pid_t));
	
	if (tmp == NULL){
		syslog(LOG_ERR, "error while trying to create pool element: %s", sys_errlist[errno]);
		return 1;
	}
	
	if (chld_st == NULL){
		chld_st = tmp;
		chld_st->pid = pid;
		chld_st->next = NULL;
		chld_st->prev = NULL;
		return 0;
	} else {
		s = chld_st;
		/* look for the last node */
		while (s->next != NULL){
			s = s->next;
		}
		s->next = tmp;
		tmp->prev = s;
		tmp->next = NULL;
		tmp->pid = pid;
		return 0;
	}
	return 1;
}


int rem_pool(pid_t pid){
	struct child_pid_t * tmp;
	syslog(LOG_INFO, "removing %i from pool.", pid);
	if (chld_st == NULL){
		return 1;
	}
	tmp = chld_st;
	while (tmp != NULL){
		/* printf(">> tmp->pid[%i] == pid[%i]\n", tmp->pid, pid); */
		if (tmp->pid == pid){
			/* remove this node*/
			if (tmp->prev == NULL){
				if (tmp->next == NULL){
					/* one element in list */
					chld_st = NULL;
					free(tmp);
					return 0;
				}
				(tmp->next)->prev = NULL;
				chld_st = tmp->next;
			} else {
				(tmp->prev)->next = tmp->next;
			}
			if (tmp->next == NULL){
				(tmp->prev)->next = NULL;
			} else {
				(tmp->next)->prev = tmp->prev;
			}
			free(tmp);
			return 0;
		}
		tmp = tmp->next;
	}
	return 1;
}


