/*
 * Copyright (C) 2004 Andres Sajo
 *
 * This file is part of the nisperod package.
 *
 * nisperod is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
 *
 *
 * $Id: parse.c,v 1.12 2004/09/30 13:09:50 andres Exp andres $ 
 */
 
#include "nisperod.h"

/* for parse_config() */
struct info_t info;
int usercount = 0;

/*
 *   parse the define table map file
 */


int parse_table(const char * table_path){

	char buff[BUFFER-1];
	char token[BUFFER];
	char * ptr = buff;
	FILE * fd;
	int line = 0, outlets = 0;



	/* table of defines */
	struct def_tab_t * def = NULL;
	struct def_tab_t * ind = NULL;
	struct def_tab_t * def_cid = NULL; /* for nis -> def seeks */

	/* table of devices */
	struct nispero_t * nis = NULL;
	struct nispero_t * tbl = NULL;
	
	defl_st = NULL;
	nisl_st = NULL;

	ind = create_def_tab_elem(NULL);
	if ( ind == NULL ) {
		perror("define");
		return EXIT_FAILURE;
	}

	nis = create_nispero_t_elem(NULL);
	if (nis == NULL) {
		perror("nispero_t");
		return EXIT_FAILURE;
	}

	fd = fopen(table_path, "ro");
	if (fd == NULL) { perror(table_path); return EXIT_FAILURE; }
	while (fgets(buff, sizeof(buff), fd) != NULL){
		line++;
		if ((strlen(buff) == sizeof(buff)-1) && buff[sizeof(buff)] != '\n'){
			fprintf(stderr, "line %i is too long.\n", line);
			/*return EXIT_FAILURE;*/
			goto e_exit;
		}
		ptr = buff;
		/* get first token */
		ptr = gettoken(ptr, token);
		if (found == 0) continue;
		if (token[0] == '#'){
			/* ignore comments */
			continue;
		} else if (strcmp(token, (char *)"define") == 0) {	
			/* it is a define need two more tokens */
			ptr = gettoken(ptr, token);
			if (found == 0) {
				fprintf(stderr, "in line %i, missing 2nd token.\n", line);
				/*return EXIT_FAILURE;*/
				goto e_exit;
			}
			if (strlen(token) > sizeof(ind->name)-1) {
				fprintf(stderr, "in line %i, token too long (%i), name > %i char.\n",  line,strlen(token), sizeof(ind->name)-1);
				/*return EXIT_FAILURE;*/
				goto e_exit;
			}
			strncpy(ind->name, token, strlen(token)+1);
			ptr = gettoken(ptr, token);
			if ( found == 0) {
				fprintf(stderr, "in line %i, missing 3rd token.\n", line);
				/*return EXIT_FAILURE;*/
				goto e_exit;
			}
			if (strlen(token) > sizeof(ind->cid)-1) {
				fprintf(stderr, "in line %i, token too long (%i), cid > %i char.\n", line, strlen(token), sizeof(ind->cid)-1);
				/*return EXIT_FAILURE;*/
				goto e_exit;
			}
			strncpy(ind->cid, token, strlen(token)+1);
			ptr = gettoken(ptr, token);
			if (ptr != NULL) {
				fprintf(stderr, "in line %i, unknown extra token.", line);
				/*return EXIT_FAILURE;*/
				goto e_exit;
			}
			if (defl_st == NULL){
				defl_st = create_def_tab_elem(NULL);
				def = defl_st;
			} else {
				def->next = create_def_tab_elem(NULL);
				def = def->next;
			}
			strcpy(def->name, ind->name);
			strcpy(def->cid, ind->cid);
			/*printf("adding def table: def->name=\'%s\', def->cid=\'%s\'\n", def->name, def->cid);*/
		} else {
			/* needs at least 3 tokens */
			/* HERE */
			if (strlen(token) > sizeof(nis->name) -1) {
				fprintf(stderr, "in line %i, token too long (%i), name > %i char.\n", line, strlen(token), sizeof(nis->name) -1);
				/*return EXIT_FAILURE;*/
				goto e_exit;
			}
			strcpy(nis->name, token);
			
			/* get the define */
			ptr = gettoken(ptr, token);
			if (found == 0) {
				fprintf(stderr, "in line %i, missing 2nd token.\n", line);
				/*return EXIT_FAILURE;*/
				goto e_exit;
			}
			def_cid = seek_def(token, defl_st);
			if (def_cid == NULL){
				fprintf(stderr, "in line %i, name '%s' not previously defined\n", line, token);
				/*return EXIT_FAILURE;*/
				goto e_exit;
			}
			sscanf(&def_cid->cid[0], "%Lx", &nis->c_id);
			/* get the outlets */
			ptr = gettoken(ptr, token);
			if (found == 0) {
				printf("in line %i, missing 3rd token.\n", line);
				/*return EXIT_FAILURE;*/
				goto e_exit;
			}
			outlets = BIT(atoi(token));
			if ((outlets & 127) == 0){ /* zero or bigger that 6 bit */
				printf("1in line %i, 3rd token not a valid number [1-6].\n", line);
				/*return EXIT_FAILURE;*/
				goto e_exit;
			}
			/* reset nis->outlets */
			nis->outlets = 0;
			nis->outlets |= outlets;
			ptr = gettoken(ptr, token);
			while ( found != 0 ){
				outlets = BIT(atoi(token));
				if ((outlets & 127) == 0){ /* zero or bigger that 6 bit */
					printf("2in line %i, token not a valid number [1-6].\n", line);
					/* return EXIT_FAILURE;*/
					goto e_exit;
				}
				nis->outlets |= outlets;
				ptr = gettoken(ptr, token);
			}
			/* copy all the above to tbl */
			if (nisl_st == NULL){
				/*printf("in defl_st == NULL\n");*/
				nisl_st = create_nispero_t_elem(NULL);
				tbl = nisl_st;
			} else {
				/*printf("in defl_st != NULL\n");*/
				tbl->next = create_nispero_t_elem(NULL);
				tbl = tbl->next;
			}
			strcpy(tbl->name, nis->name);
			tbl->c_id = nis->c_id;
			tbl->outlets = nis->outlets;
		}
	}
	fclose(fd);
	return EXIT_SUCCESS;
e_exit:
	fclose(fd);
	return EXIT_FAILURE;
}



/*
 *    parse the config file
 */
int parse_config (const char * config_path){
	
	char buff[BUFFER-1];
	char token[BUFFER];
	char * ptr = buff;
	FILE * fd;
	int line = 0;
	usercount = -1;
	
	fd = fopen(config_path, "ro");
	
	if (fd == NULL) { perror(config_path); return EXIT_FAILURE; }
	while (fgets(buff, sizeof(buff), fd) != NULL){
		line++;
		if ((strlen(buff) == sizeof(buff)-1) && buff[sizeof(buff)] != '\n'){
			fprintf(stderr, "line %i is too long.\n", line);
			/*return EXIT_FAILURE;*/
			goto e_exit;
		}
		ptr = buff;
		/* get first token */
		ptr = gettoken(ptr, token);
		if (found == 0) continue;
		if (token[0] == '#'){
			/* ignore comments */
			continue;
		/* first case : port */
		} else if (strcmp(token, (char *)"port") == 0) {
			/*printf("in port\n");*/
			/* it is a define need one more token */
			ptr = gettoken(ptr, token);
			if (found == 0) {
				fprintf(stderr, "in line %i, missing 2nd token.\n", line);
				/*return EXIT_FAILURE;*/
				goto e_exit;
			}
			if (strlen(token) > sizeof(info.port)-1) {
				fprintf(stderr, "in line %i, token too long (%i), port > %i char.\n",  line, strlen(token), sizeof(info.port)-1);
				/*return EXIT_FAILURE;*/
				goto e_exit;
			}
			strncpy(info.port, token, strlen(token)+1);
			ptr = gettoken(ptr, token);
			if ( ptr != NULL) {
				fprintf(stderr, "in line %i, unknown extra 3rd token.\n", line);
				/*return EXIT_FAILURE;*/
				goto e_exit;
			}
		} else if (strcmp(token, (char *)"socket") == 0) {
			/*printf("in socket\n");*/
			/* it is a define need one more token */
			ptr = gettoken(ptr, token);
			if (found == 0) {
				fprintf(stderr, "in line %i, missing 2nd token.\n", line);
				/*return EXIT_FAILURE;*/
				goto e_exit;
			}
			if (strlen(token) > sizeof(info.socket)-1) {
				fprintf(stderr, "in line %i, token too long (%i), port > %i char.\n",  line, strlen(token), sizeof(info.port)-1);
				/*return EXIT_FAILURE;*/
				goto e_exit;
			}
			strncpy(info.socket, token, strlen(token)+1);
			ptr = gettoken(ptr, token);
			if ( ptr != NULL) {
				fprintf(stderr, "in line %i, unknown extra 3rd token.\n", line);
				/*return EXIT_FAILURE;*/
				goto e_exit;
			}
		} else if (strcmp(token, (char *)"auth") == 0) {
			/* printf("in auth\n"); */
			usercount++;
			if (usercount > MAX_USERS){
				fprintf(stderr, "in line %i, max number of auth reached (%i).\n", line, MAX_USERS);
			}
			/* it is a define need three more tokens */
			ptr = gettoken(ptr, token);
			if (found == 0) {
				fprintf(stderr, "in line %i, missing 2nd token.\n", line);
				/*return EXIT_FAILURE;*/
				goto e_exit;
			}
			if (strlen(token) > sizeof(info.users[usercount].login)-1) {
				fprintf(stderr, "in line %i, token too long (%i), login > %i char.\n",  line, strlen(token), sizeof(info.users[usercount].login)-1);
				/*return EXIT_FAILURE;*/
				goto e_exit;
			}
			strncpy(info.users[usercount].login, token, strlen(token)+1);
			ptr = gettoken(ptr, token);
			/* password */
			if (found == 0) {
				fprintf(stderr, "in line %i, missing 3rd token.\n", line);
				/*return EXIT_FAILURE;*/
				goto e_exit;
			}
			if (strlen(token) > sizeof(info.users[usercount].password)-1) {
				fprintf(stderr, "in line %i, token too long (%i), password > %i char.\n",  line, strlen(token), sizeof(info.users[usercount].password)-1);
				/*return EXIT_FAILURE;*/
				goto e_exit;
			}
			strncpy(info.users[usercount].password, token, strlen(token)+1);
			ptr = gettoken(ptr, token);
			/* protocol type */
			if (found == 0) {
				fprintf(stderr, "in line %i, missing 4th token.\n", line);
				/*return EXIT_FAILURE;*/
				goto e_exit;
			}
			if (strcmp(token, (char *)"extended") == 0) {
				info.users[usercount].protocol = PROTO_EXT;
			} else if (strcmp(token, (char *)"simple") == 0) {
				info.users[usercount].protocol = PROTO_SIM;
			} else {
				fprintf(stderr, "in line %i, 4th token unknown protocol type\n", line);
				goto e_exit;
			}
			ptr = gettoken(ptr, token);
			/* should end here */
			if ( ptr != NULL) {
				fprintf(stderr, "in line %i, unknown extra 3rd token.\n", line);
				/*return EXIT_FAILURE;*/
				goto e_exit;
			}
		} else {
				fprintf(stderr, "in line %i, unknown token \'%s\'.\n", line, token);
				/*return EXIT_FAILURE;*/
				goto e_exit;
		}
	}
	fclose(fd);
	return EXIT_SUCCESS;
e_exit:
	fclose(fd);
	return EXIT_FAILURE;
}

