/* uszerver.c - Egyszeru pelda szerver a Unix domain socket 
   hasznalatara. */

#include <stdio.h>
#include <stdlib.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <unistd.h>
#include <string.h>

#define SOCKET_PATH "/tmp/stonith.socket"

#define SEND(text)	write(conn, text "\n", strlen(text)+1)
#define SENDOK		write(conn, "OK\n", 3)
#define SENDFAIL	write(conn, "INVALID\n", 8)

int main(void)
{
  struct sockaddr_un address;
  int sock, conn, login=0;
  size_t addrLength;
  char buf[1024];
  int amount;

  /* Letrehozzuk a socket-et. */
  if((sock = socket(PF_UNIX, SOCK_STREAM, 0)) < 0)
  {
    perror("socket");
    exit(1);
  }

  /* Letoroljuk a korabbi socket allomanyt. */
  unlink(SOCKET_PATH);

  address.sun_family = AF_UNIX;
  strcpy(address.sun_path, SOCKET_PATH);

  /* A teljes cim hossz tartalmazza a sun_family elemet es az 
     eleresi ut hosszat. */
  addrLength = sizeof(address.sun_family) + 
               strlen(address.sun_path);

  /* A socket-et hozzakotjuk a cimhez. */
  if(bind(sock, (struct sockaddr *) &address, addrLength))
  {
    perror("bind");
    exit(1);
  }

  /* Bekapcsoljuk a kapcsolodasra valo varakozast. */
  if(listen(sock, 5))
  {
    perror("listen");
    exit(1);
  }

  /* Fogadjuk a kapcsolodasokat. */
  while((conn = accept(sock, (struct sockaddr *) &address, 
    &addrLength)) >= 0)
  {
    /* Fogadjuk az adatokat. */
    printf("Link established...\n");
    login=0;
    while ((amount = read(conn, buf, sizeof(buf))) > 0) 
    {
    	buf[amount] = '\0';
	printf("---> '%s'\n", buf);    
	if ( strncmp(buf, "REGISTER:", 9) == 0 ){
		if ( strncmp(buf+9, "STONITH\n", 8) == 0 ) {
			SENDOK;
			login=1;
		} else {
			SENDFAIL;
		}
		
	} else if ( strncmp(buf, "PASSWORD:", 9) == 0 ){
		if ( (strncmp(buf+9, "STONITH\n", 8) == 0) && (login == 1 )) {
			SENDOK;
			login=2;
		} else {
			SENDFAIL;
		}
	} else if ( ( strncmp(buf, "STATUS:", 7) == 0 ) && (login == 2)){
		SENDOK;
	} else if ( ( strncmp(buf, "ON:", 3) == 0 ) && (login == 2)){
		SENDOK;
	} else if ( (strncmp(buf, "OFF:", 4) == 0 ) && (login == 2)){
		SENDOK;
	} else {
		SENDFAIL;
	}
    }
    if(amount < 0)
    {
      perror("read");
      exit(1);
    }
    printf("Link closed\n");

    /* Bontjuk a kapcsolatot. */
    close(conn);
  }

  if(conn < 0) 
  {
    perror("accept");
    exit(1);
  }

  /* Lezarjuk a szerver socket-et. */    
  close(sock);
  return 0;
}
