/*
 * Copyright (C) 2004 Andres Sajo
 *
 * This file is part of the nisperod package.
 *
 * nisperod is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
 *
 *
 *$Id: unixds.c,v 1.9 2004/10/08 14:31:24 andres Exp andres $
 */

/* REF[1] */
#include <sys/socket.h>
#include <sys/un.h>
#include "nisperod.h"

#define SEND(text)	write(nis_conn, text "\n", strlen(text)+1)
#define SENDVAR(var)	write(nis_conn, tmp, strlen(var))
#define SENDOK		write(nis_conn, "OK\n", 3)
#define SENDFAIL	write(nis_conn, "INVALID\n", 8)

int register_connect(); /* returns the user id or -1 on error */

struct sockaddr_un nis_address;
int nis_sock;
size_t nis_addrLength;

int open_socket(void){ /* error can go to stderr, here we are not wet backgounded */
	if((nis_sock = socket(PF_UNIX, SOCK_STREAM, 0)) < 0){
		perror("socket");
		exit(1);
	}

	/* Letoroljuk a korabbi socket allomanyt. */
	unlink(info.socket);

	nis_address.sun_family = AF_UNIX;
	strcpy(nis_address.sun_path, info.socket);

	/* A teljes cim hossz tartalmazza a sun_family elemet es az 
	 * eleresi ut hosszat. 
	 */
	nis_addrLength = sizeof(nis_address.sun_family) + 
				strlen(nis_address.sun_path);

	/* A socket-et hozzakotjuk a cimhez. */
	if(bind(nis_sock, (struct sockaddr *) &nis_address, nis_addrLength)){
		perror("bind");
		return 1;
	}

	/* Bekapcsoljuk a kapcsolodasra valo varakozast. */
	/* don't change this value, access should be semaphored and synchronized:
	 * that is, if ``listen'' is greater than 1, access would be asynchronious
	 * and the code would break.
	 */
	if(listen(nis_sock, 1)){
		perror("listen");
		return 1;
	}
  
	return 0;
}

/* for child signal handling process */
#include <signal.h>
/* #include <syslog.h> */
/* this code alwais executes has child */
/* we don't use term here! */
int nis_conn;
int accept_connect(void){ /* errors go to syslog(LOG_ERROR, ... */
	char nis_buf[BUFFER];
	char tmp[BUFFER];
	int nis_amount;
	int cu = 0, i = 0;
	char * p;
	struct nispero_t * device = NULL;
	struct def_tab_t * def = NULL;
	struct nispero_t * nis = NULL;
	device = NULL;
	/* Fogadjuk a kapcsolodasokat. */
	/*printf("we are in accept_connect(%i, %p, %p);\n", nis_sock, &nis_address, &nis_addrLength);*/
	
	nis_conn = accept(nis_sock, (struct sockaddr *) &nis_address,
						 &nis_addrLength);
	if (nis_conn < 0){
		syslog(LOG_ERR, "child: error opening connection, %i: %s.",
							 errno, sys_errlist[errno]);
		return 1;	
	}
	
	/*syslog(LOG_ERR, "*** child, nis_conn=%i", nis_conn);*/
	syslog(LOG_INFO, "child: installing own termination handler");
	
	if (signal(SIGUSR1, &unixds_term) == SIG_ERR){
		syslog(LOG_ERR,
			"child: unable to hook signal SIGUSR1 to handler &unixds_term().");
		return 1;
	}
	
	/*printf("Link established...\n");*/
	syslog(LOG_INFO, "child: connection established.");
	cu = register_connect();
	if ( cu < 0){
		syslog(LOG_INFO, "child: couldn't authentificate connection.");
		close_accept_connect();
		return 1;
	}
	syslog(LOG_INFO, "child: user %s authentificated, using the %s protocol.",
			info.users[cu].login, PRTY(info.users[cu].protocol));
	
	/* registerd ! with potocol info.users[cu].protocol */
	
	if ( info.users[cu].protocol == PROTO_SIM ){
		/* 
		 * Using protocol version 1.1
		 * the simple protocol parser
		 * mostly for use with heartbeat
		 */
		while ((nis_amount = read(nis_conn, nis_buf, sizeof(nis_buf))) > 0){
			nis_buf[nis_amount] = '\0';
			/* printf("--> parsing: \'%s\'", nis_buf); */
			/*
			 *    STATUS ******************************
			 */
			if ( strncmp(nis_buf, "STATUS:", 7) == 0 ){
				if (wrapper_open_dev(info.port) != 0){
					syslog(LOG_ALERT, "child: bus open failed, is it plugged?");
					SENDFAIL;
				}
				if ( wrapper_status() != 0 ){
					syslog(LOG_ALERT, "child: bus reset failed, is it plugged?");
					SENDFAIL;
				} else {
					syslog(LOG_INFO, "child: bus reset ok.");
					SENDOK;
				}
				wrapper_close_dev();
			/*
			 *    ON **************************************
			 */
			} else if ( strncmp(nis_buf, "ON:", 3) == 0 ) {
				/* printf("--> ON: \'%s\'", nis_buf); */
				p = gettoken(nis_buf+3, tmp);
				if (found == 0){
					SENDFAIL;
					continue;
				} else {
					gettoken(p, tmp);
					if (found != 0){ /* there is a third token ... */
						SENDFAIL;
						continue;
					}
				}
				/* printf("--> wrap_turn_on(%s)\n", tmp); */
				if (wrapper_open_dev(info.port) != 0){
					syslog(LOG_ALERT, "child: bus open failed, is it plugged?");
					SENDFAIL;
				}
				if (wrapper_turn_on(tmp) != 0){
					syslog(LOG_INFO, "child: ON %s: failed.", tmp);
					SENDFAIL;
				} else {
					syslog(LOG_INFO, "child: ON %s: ok.", tmp);
					SENDOK;
				}
				wrapper_close_dev();
			/*
			 *    OFF *************************************
			 */
			} else if (strncmp(nis_buf, "OFF:", 4) == 0 ) {
				p = gettoken(nis_buf+4, tmp);
				if (found == 0){
					SENDFAIL;
					continue;
				} else {
					gettoken(p, tmp);
					if (found != 0){ /* there is a third token ... */
						SENDFAIL;
						continue;
					}
				}
				if (wrapper_open_dev(info.port) != 0){
					syslog(LOG_ALERT, "child: bus open failed, is it plugged?");
					SENDFAIL;
				}
				if (wrapper_turn_off(tmp) != 0){
					syslog(LOG_INFO, "child: OFF %s: failed.", tmp);
					SENDFAIL;
				} else {
					syslog(LOG_INFO, "child: OFF %s: ok.", tmp);
					SENDOK;
				}
				wrapper_close_dev();
			/*
			 *    else *********************************
			 */
			} else {
				SENDFAIL;
			}
		}
	} else {
		/* the extended protocol parser
		 * this is the protocol humans should use ...
		 */
		while ((nis_amount = read(nis_conn, nis_buf, sizeof(nis_buf))) > 0){
			nis_buf[nis_amount] = '\0';
			/*
			 *    reset ******************************
			 */
			p = gettoken(nis_buf, tmp);
			if (found == 0){
				SEND("");
				continue;
			}
			if ( strcmp(tmp, "reset") == 0 ){
				gettoken(p, tmp);
				if (found != 0){
					SENDFAIL;
					continue;
				}
				if (wrapper_open_dev(info.port) != 0){
					syslog(LOG_ALERT, "child: bus open failed, is it plugged?");
					SENDFAIL;
				}
				if ( wrapper_status() != 0 ){
					syslog(LOG_ALERT, "child: bus reset failed, is it plugged?");
					SENDFAIL;
				} else {
					syslog(LOG_INFO, "child: bus reset ok.");
					SENDOK;
				}
				wrapper_close_dev();
			/*
			 *    list **************************************
			 */
			} else if ( strcmp(tmp, "list") == 0 ){
				gettoken(p, tmp);
				if (found != 0){
					SENDFAIL;
					continue;
				}
				if (wrapper_open_dev(info.port) != 0){
					syslog(LOG_ALERT, "child: bus open failed, is it plugged?");
					SENDFAIL;
				}
				if ( wrapper_status() != 0 ){
					syslog(LOG_ALERT, "child: bus reset of device failed, is it plugged?");
					SENDFAIL;
					continue;
				}
				wrapper_close_dev();
				
				SEND("auth table:     login              password      protocol");
				while (i <= usercount){
					sprintf(tmp, "\t\'%-16s\'            !          %s\n",  
					info.users[i].login, PRTY(info.users[i].protocol));
					SENDVAR(tmp);
					i++;
				}
				
				SEND("\ndefine table: define name                     value");
				def = defl_st;
				while (def != NULL){
					sprintf(tmp, "\'%-32s\' -> \'%18s\'\n", def->name, def->cid);
					SENDVAR(tmp);
					def = def->next;
				}
			
				SEND("\nmapping table:  device                    chip address      outlet(s)");
				nis = nisl_st;
				while (nis != NULL){
					sprintf(tmp, "\'%-32s\' ->  0x%016Lx     0x%02x\n",
					nis->name, nis->c_id, nis->outlets);
					SENDVAR(tmp);
					nis = nis->next;
				}
				SENDOK;
			/*
			 *    get  **************************************
			 */
			} else if ( strcmp(tmp, "get") == 0 ){
				p = gettoken(p, tmp);
				if (found == 0){
					SENDFAIL;
					continue;
				}
				/* look for the device address */
				def = seek_def(tmp, defl_st);
				if (def == NULL){
					SENDFAIL;
					continue;
				}
				p = gettoken(p, tmp);
				if (found != 0){
					SENDFAIL;
					continue;
				}
				if (wrapper_open_dev(info.port) != 0){
					syslog(LOG_ALERT, "child: bus open failed, is it plugged?");
					SENDFAIL;
				}
				i = wrapper_get(def);
				wrapper_close_dev();
				if (i < 0) {
					SENDFAIL;
					continue;
				}
				sprintf(tmp, "%s:0x%02x\n", def->name, i);
				SENDVAR(tmp);
				SENDOK;
			/*
			 *    set  **************************************
			 */
			} else if ( strcmp(tmp, "set") == 0 ){
				/* printf("in set\n"); */
				p = gettoken(p, tmp);
				if (found == 0){
					SENDFAIL;
					continue;
				}
				/* printf("(1)\n"); */
				/* look for the device address */
				def = seek_def(tmp, defl_st);
				/* printf("(def = %p)\n", def); */
				if (def == NULL){
					SENDFAIL;
					continue;
				}
				p = gettoken(p, tmp);
				/* printf("(2)\n"); */
				if (found == 0){
					SENDFAIL;
					continue;
				}
				
				/* i = atoi(tmp); */
				sscanf(tmp, "%i", &i);
				/* printf("(i = %i)\n", i); */
				/* printf(">>> set, integer: %i\n", i); */
				if ( (i > 0xff) || (i < 0) ){
					SENDFAIL;
					continue;
				}
				p = gettoken(p, tmp);
				if (found != 0){
					SENDFAIL;
					continue;
				}
				if (wrapper_open_dev(info.port) != 0){
					syslog(LOG_ALERT, "child: bus open failed, is it plugged?");
					SENDFAIL;
				}
				i = wrapper_set(def, i);
				wrapper_close_dev();
				if (i < 0) {
					SENDFAIL;
					continue;
				}
				sprintf(tmp, "%s:0x%02x\n", def->name, i);
				SENDVAR(tmp);
				SENDOK;
			/*
			 *    else **************************************
			 */
			} else {
				SENDFAIL;
			}
		}
/*
		SEND("protocol not yet implemented");
		SENDFAIL;
		*/
	}
	
#if 0
	while ((nis_amount = read(nis_conn, nis_buf, sizeof(nis_buf))) > 0){
		nis_buf[nis_amount] = '\0';
		printf("child(%i) -> '%s'", getpid(), nis_buf);
		if ( strncmp(nis_buf, "REGISTER:", 9) == 0 ){
			/* Get username */
			if ( strncmp(nis_buf+9, "STONITH\n", 8) == 0 ) {
				SENDOK;
				nis_login=1;
			} 
		}
		
		if (nis_login != 1){
			SENDFAIL;
			return 1;
		}
		
		if ( strncmp(nis_buf, "PASSWORD:", 9) == 0 ){
			if ( (strncmp(nis_buf+9, "STONITH\n", 8) == 0) && (nis_login == 1 )) {
				SENDOK;
				nis_login=2;
			} else {
				SENDFAIL;
			}
		} else if ( ( strncmp(nis_buf, "STATUS:", 7) == 0 ) && (nis_login == 2)){
			SENDOK;
		} else if ( ( strncmp(nis_buf, "ON:", 3) == 0 ) && (nis_login == 2)){
			SENDOK;
		} else if ( (strncmp(nis_buf, "OFF:", 4) == 0 ) && (nis_login == 2)){
			SENDOK;
		} else {
			SENDFAIL;
		}
	}
#endif
/*	printf("child(%i), term = %i\n", getpid(), term); */
	return close_accept_connect();
}

int close_accept_connect(void){
	/* give the parent some time to breath: */
	/* sleep(1); */
	close(nis_conn);
	if(nis_conn < 0) {
		syslog(LOG_ERR, "child: error closing connection, %i: %s.", errno, sys_errlist[errno]);
		return 1;
	}
	syslog(LOG_INFO, "child: connection closed.");
	/*printf("Link closed\n");*/
	return 0;
}

int register_connect(void){
	int nis_amount = 0, i = 0;
	int cu = -1;
	char s[BUFFER];
	char * b;
	const char * pr;
#if 0
	syslog(LOG_INFO, ">>> child: nis_conn=%i.", nis_conn);
#endif
	if ((nis_amount = read(nis_conn, s, sizeof(s))) > 0){
		s[nis_amount-1] = '\0';
#if 0
		syslog(LOG_INFO, ">>> child(%i) 1: REGISTER -> %p:'%s'", getpid(), s, s);
#endif
		pr = REGISTER;
		if ( strncmp(s, pr, L_REGISTER) == 0 ){
			/* Get username */
			b = s+L_REGISTER;
#if 0
			syslog(LOG_INFO, ">>> child: in compare. [%s]", b);
#endif
			for (i = 0; i <= usercount; i++){
#if 0
				syslog(LOG_INFO, "'%s' == '%s'\n", b, info.users[i].login);
#endif
				if ( strcmp(b, info.users[i].login) == 0 ) {
					SENDOK;
					cu=i;
					goto verify_password;
				}
			}
		}
	}
	SENDFAIL;
	return -1;

verify_password:
	if ((nis_amount = read(nis_conn, s, sizeof(s))) > 0){
		s[nis_amount-1] = '\0';
#if 0
		syslog(LOG_INFO, ">>> child(%i) 1: PASSWORD -> %p:'%s'", getpid(), s, s);
#endif
		pr = PASSWORD;
		if ( strncmp(s, pr, L_PASSWORD) == 0 ){
			/* Get password */
			b = s + L_PASSWORD;
#if 0
			syslog(LOG_INFO, ">>> child: in compare. [%s]", b);
			syslog(LOG_INFO, ">>> '%s' == '%s'\n", b, info.users[i].password);
#endif
			if ( strcmp(b, info.users[cu].password) == 0 ) {
				SENDOK;
				return cu;
			}
		}
	}
	SENDFAIL;
	return -1;
}
